/* 
 * File:   gdgpart.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 13. května 2019, 22:39
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdint.h>
#include <glib.h>

#include "gdgelement.h"
#include "gdgpart.h"
#include "main.h"


static void gdgpart_destroy ( st_GDGPART *part ) {
    if ( !part ) return;
    if ( part->netpin ) g_free ( part->netpin );
    g_free ( part );
}


void gdgparts_destroy ( st_GDGPARTSALL *partall ) {
    if ( !partall ) return;

    if ( partall->type == GDGPARTSALL_TYPE_MASTER ) {
        int i;
        for ( i = 0; i < partall->count_part; i++ ) {
            st_GDGPART *part = partall->part[i];
            gdgpart_destroy ( part );
        };

        if ( partall->part ) g_free ( partall->part );
    };

    if ( partall->loc ) g_free ( partall->loc );
    g_free ( partall );
}


static void gdgparts_add_part ( st_GDGPARTSALL *partall, st_GDGELEMENT *elm, int loc_x, int loc_y ) {

    st_GDGPART *part = g_new0 ( st_GDGPART, 1 );
    part->element = elm;
    part->loc_x = loc_x;
    part->loc_y = loc_y;
    part->netpin = g_new0 ( st_GDGNET*, elm->count_port );

    int i = partall->count_part;
    partall->count_part++;
    if ( !i ) {
        partall->part = g_new0 ( st_GDGPART*, 1 );
    } else {
        partall->part = g_renew ( st_GDGPART*, partall->part, partall->count_part );
    };
    partall->part[i] = part;

    if ( part->loc_x < partall->min_loc_x ) partall->min_loc_x = part->loc_x;
    if ( part->loc_y < partall->min_loc_y ) partall->min_loc_y = part->loc_y;
    if ( part->loc_x > partall->max_loc_x ) partall->max_loc_x = part->loc_x;
    if ( part->loc_y > partall->max_loc_y ) partall->max_loc_y = part->loc_y;

    gdgelement_register_part_instance ( partall->elmall, elm, part );
}


static int gdgparts_get_real_part_location ( st_GDGPARTSALL *partall, int loc_x, int loc_y ) {
    int real_x = loc_x - partall->min_loc_x;
    int real_y = loc_y - partall->min_loc_y;
    return real_y + ( real_x * partall->ysize );
}


static gboolean gdgparts_reindex_part_locations ( st_GDGPARTSALL *partall ) {
    if ( partall->loc ) g_free ( partall->loc );
    partall->loc = NULL;
    if ( partall->count_part ) {
        partall->xsize = ( partall->max_loc_x - partall->min_loc_x ) + 1;
        partall->ysize = ( partall->max_loc_y - partall->min_loc_y ) + 1;
        int grid_size = partall->xsize * partall->ysize;
        /*
                printf ( "Locations grid size: %d\n", grid_size );
                printf ( "\tx: <%d, %d>\n", partall->min_loc_x, partall->max_loc_x );
                printf ( "\ty: <%d, %d>\n", partall->min_loc_y, partall->max_loc_y );
         */
        partall->loc = g_new0 ( st_GDGPART*, grid_size );
        int i;
        for ( i = 0; i < partall->count_part; i++ ) {
            st_GDGPART *part = partall->part[i];
            int real_loc = gdgparts_get_real_part_location ( partall, part->loc_x, part->loc_y );
            if ( partall->loc[real_loc] ) {
                fprintf ( stderr, "%s(%d) - ERROR: %s_%d_%d - can't be registered, place is used by %s\n", part->element->name, part->loc_x, part->loc_y, partall->loc[real_loc]->element->name );
                return FALSE;
            };
            partall->loc[real_loc] = part;
        };
    };
    return TRUE;
}


st_GDGPARTSALL* gdgparts_init ( st_GDGELEMENTSALL *elmall, const char *filepath ) {

    FILE *fp = fopen ( filepath, "r" );
    if ( !fp ) {
        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "%s(%d) - ERROR: Can't open input file '%s'\n", __func__, __LINE__, filepath );
        main_error_exit ( g_string_free ( str, FALSE ) );
        return NULL;
    };

    st_GDGPARTSALL *partall = g_new0 ( st_GDGPARTSALL, 1 );
    partall->min_loc_x = 9999;
    partall->min_loc_y = 9999;
    partall->type = GDGPARTSALL_TYPE_MASTER;
    partall->elmall = elmall;

    char buf[1024];

    GRegex *regex = g_regex_new ( "^x(F\\d+)_(\\d+)_(\\d+)\\s", 0, 0, NULL );

    while ( fgets ( buf, sizeof ( buf ), fp ) ) {
        GMatchInfo *match_info;

        g_regex_match ( regex, buf, 0, &match_info );
        while ( g_match_info_matches ( match_info ) ) {
            gchar *elm_name = g_match_info_fetch ( match_info, 1 );
            gchar *loc_x_txt = g_match_info_fetch ( match_info, 2 );
            gchar *loc_y_txt = g_match_info_fetch ( match_info, 3 );
            int loc_x = atoi ( loc_x_txt );
            int loc_y = atoi ( loc_y_txt );

            st_GDGELEMENT *elm = gdgelements_search_element_by_name ( elmall, elm_name );
            if ( !elm ) {
                fclose ( fp );
                return NULL;
            };

            //g_print ( "Found: %s - x: %d, y: %d\n", elm_name, loc_x, loc_y );

            gdgparts_add_part ( partall, elm, loc_x, loc_y );

            g_free ( elm_name );
            g_free ( loc_x_txt );
            g_free ( loc_y_txt );
            g_match_info_next ( match_info, NULL );
        }
        g_match_info_free ( match_info );
    };

    g_regex_unref ( regex );

    fclose ( fp );

    fprintf ( stderr, "%s() - initialised %d parts\n", __func__, partall->count_part );

    if ( !gdgparts_reindex_part_locations ( partall ) ) return NULL;

    return partall;
}


st_GDGPART* gdgparts_get_part_by_location ( st_GDGPARTSALL *partall, int loc_x, int loc_y, gboolean print_error ) {
    if ( ( loc_x < partall->min_loc_x ) || ( loc_x > partall->max_loc_x ) || ( loc_y < partall->min_loc_y ) || ( loc_y > partall->max_loc_y ) ) {
        if ( print_error ) {
            fprintf ( stderr, "%s(%d) - Error: invalid part location x: %d, y: %d\n", __func__, __LINE__, loc_x, loc_y );
        };
        return NULL;
    };
    int real_loc = gdgparts_get_real_part_location ( partall, loc_x, loc_y );
    return partall->loc[real_loc];
}


char* gdgpart_create_name ( st_GDGPART *part ) {
    GString *str = g_string_new ( "" );
    g_string_append_printf ( str, "%s_%d_%d", part->element->name, part->loc_x, part->loc_y );
    return g_string_free ( str, FALSE );
}


st_GDGPARTSALL* gdgparts_create_slave_from_parts ( st_GDGPART **parts, int count_parts, st_GDGELEMENTSALL *elmall ) {
    if ( !count_parts ) return NULL;

    st_GDGPARTSALL *partall = g_new0 ( st_GDGPARTSALL, 1 );
    partall->min_loc_x = 9999;
    partall->min_loc_y = 9999;
    partall->type = GDGPARTSALL_TYPE_SLAVE;

    partall->count_part = count_parts;
    partall->part = parts;
    partall->elmall = elmall;

    if ( !gdgparts_reindex_part_locations ( partall ) ) return NULL;

    return partall;
}


void gdgparts_remove_part ( st_GDGELEMENTSALL *elmall, st_GDGPARTSALL *partall, st_GDGPART *remove_part ) {
    int i;
    gboolean search = TRUE;
    for ( i = 0; i < partall->count_part; i++ ) {
        st_GDGPART *part = partall->part[i];
        if ( search ) {
            if ( remove_part == part ) {
                search = FALSE;
                gdgelement_unregister_part_instance ( elmall, part->element, remove_part );
                gdgpart_destroy ( part );
            };
        } else {
            partall->part[( i - 1 )] = part;
        };
    };

    if ( !search ) {
        partall->count_part--;
        partall->part = g_renew ( st_GDGPART*, partall->part, partall->count_part );
    };
}


void gdgparts_replace_net ( st_GDGPARTSALL *partall, st_GDGNET *oldnet, st_GDGNET *newnet ) {
    int i;
    for ( i = 0; i < partall->count_part; i++ ) {
        st_GDGPART *part = partall->part[i];
        int j;
        for ( j = 0; j < part->element->count_port; j++ ) {
            if ( part->netpin[j] == oldnet ) {
                part->netpin[j] = newnet;
            };
        };
    };
}
