/* 
 * File:   gdgpath.c
 * Author: Michal Hucik <hucik@ordoz.com>
 *
 * Created on 21. května 2019, 11:19
 * 
 * 
 * ----------------------------- License -------------------------------------
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <glib.h>

#include "gdgelement.h"
#include "gdgpart.h"
#include "gdgnet.h"
#include "main.h"

#include "gdgpath.h"
#include "vhdl.h"


void gdgpath_destroy ( st_GDGPATH *path ) {
    if ( !path ) return;
    if ( path->part ) g_free ( path->part );
    if ( path->net ) g_free ( path->net );
    if ( path->pin ) g_free ( path->pin );
    if ( path->element ) g_free ( path->element );
    if ( path->str ) g_string_free ( path->str, TRUE );
    g_free ( path );
}


st_GDGPATH* gdgpath_new ( st_GDGPARTSALL *partall ) {
    st_GDGPATH *path = g_new0 ( st_GDGPATH, 1 );
    path->pin = g_new0 ( st_GDGNET_PIN*, 1 );
    path->net = g_new0 ( st_GDGNET*, 1 );
    path->part = g_new0 ( st_GDGPART*, 1 );
    path->element = g_new0 ( st_GDGELEMENT*, 1 );
    path->partall = partall;
    path->create_path_str = FALSE;
    path->verbose = FALSE;
    path->str = g_string_new ( "" );
    return path;
}


int gdgpath_path_has_net ( st_GDGPATH *path, st_GDGNET *net ) {
    int i;
    for ( i = 0; i < path->net_count; i++ ) {
        if ( path->net[i] == net ) return i;
    };
    return -1;
}


int gdgpath_path_has_pin ( st_GDGPATH *path, st_GDGNET_PIN *pin ) {
    int i;
    for ( i = 0; i < path->pin_count; i++ ) {
        if ( path->pin[i] == pin ) return i;
    };
    return -1;
}


int gdgpath_path_has_part ( st_GDGPATH *path, st_GDGPART *part ) {
    int i;
    for ( i = 0; i < path->part_count; i++ ) {
        if ( path->part[i] == part ) return i;
    };
    return -1;
}


int gdgpath_path_has_element ( st_GDGPATH *path, st_GDGELEMENT *elm ) {
    int i;
    for ( i = 0; i < path->element_count; i++ ) {
        if ( path->element[i] == elm ) return i;
    };
    return -1;
}


gboolean gdgpath_add_net ( st_GDGPATH *path, st_GDGNET *net ) {
    if ( -1 != gdgpath_path_has_net ( path, net ) ) return TRUE;

    int i = path->net_count++;
    path->net = g_renew ( st_GDGNET*, path->net, path->net_count );
    path->net[i] = net;

    if ( net->is_constant ) {

        GString *str = g_string_new ( "" );
        g_string_append_printf ( str, "OK, net source is real pin: %s\n", net->name );
        if ( path->create_path_str ) {
            g_string_append ( path->str, str->str );
        };
        if ( path->verbose ) {
            fprintf ( stderr, "%s", str->str );
        };
        g_string_free ( str, TRUE );

        return TRUE;
    };

    st_GDGNET_PIN *pin = gdgnet_get_net_source_pin ( net );
    if ( !pin ) {
        fprintf ( stderr, "%s(%d) - not found source pin for net: %s!\n", __func__, __LINE__, net->name );
        return FALSE;
    } else {
        if ( pin->element ) {

            GString *str = g_string_new ( "OK, net source is part\n" );
            if ( path->create_path_str ) {
                g_string_append ( path->str, str->str );
            };
            if ( path->verbose ) {
                fprintf ( stderr, "%s", str->str );
            };
            g_string_free ( str, TRUE );

            st_GDGPART *part = gdgparts_get_part_by_location ( path->partall, pin->loc_x, pin->loc_y, TRUE );
            if ( !part ) {
                fprintf ( stderr, "%s(%d) - part not found: %s_%d_%d!\n", __func__, __LINE__, pin->element->name, pin->loc_x, pin->loc_y );
                return FALSE;
            } else {
                char *part_name = gdgpart_create_name ( part );
                GString *str = g_string_new ( "" );
                g_string_append_printf ( str, "\tpart name: %s\n", part_name );
                if ( path->create_path_str ) {
                    g_string_append ( path->str, str->str );
                };
                if ( path->verbose ) {
                    fprintf ( stderr, "%s", str->str );
                };
                g_string_free ( str, TRUE );
                g_free ( part_name );

                if ( !gdgpath_add_part ( path, part ) ) return FALSE;
            };
        } else {
            st_GDGNET *pinparent = pin->parent_net;

            GString *str = g_string_new ( "" );
            g_string_append_printf ( str, "OK, net source is real pin: %s (%d)\n", pinparent->name, pin->loc_z );
            if ( path->create_path_str ) {
                g_string_append ( path->str, str->str );
            };
            if ( path->verbose ) {
                fprintf ( stderr, "%s", str->str );
            };
            g_string_free ( str, TRUE );

            gdgpath_add_pin ( path, pin );
        };
    };
    return TRUE;
}


void gdgpath_add_pin ( st_GDGPATH *path, st_GDGNET_PIN *pin ) {
    if ( -1 != gdgpath_path_has_pin ( path, pin ) ) return;

    int i = path->pin_count++;
    path->pin = g_renew ( st_GDGNET_PIN*, path->pin, path->pin_count );
    path->pin[i] = pin;
}


void gdgpath_add_element ( st_GDGPATH *path, st_GDGELEMENT *elm ) {
    if ( -1 != gdgpath_path_has_element ( path, elm ) ) return;

    int i = path->element_count++;
    path->element = g_renew ( st_GDGELEMENT*, path->element, path->element_count );
    path->element[i] = elm;
}


gboolean gdgpath_add_part ( st_GDGPATH *path, st_GDGPART *part ) {
    if ( -1 != gdgpath_path_has_part ( path, part ) ) return TRUE;

    int i = path->part_count++;
    path->part = g_renew ( st_GDGPART*, path->part, path->part_count );
    path->part[i] = part;

    st_GDGELEMENT *elm = part->element;
    gdgpath_add_element ( path, elm );

    for ( i = 0; i < elm->count_port; i++ ) {
        st_GDGELEMENT_PORT *port = elm->port[i];
        if ( port->type == GDGELEMENT_PORT_IN ) {
            st_GDGNET *net = part->netpin[i];
            if ( !gdgpath_add_net ( path, net ) ) return FALSE;
        };
    };

    return TRUE;
}


st_GDGPATH* gdgpath_create_for_net ( st_GDGPARTSALL *partall, st_GDGNET *net, gboolean create_path_str, gboolean verbose ) {
    st_GDGPATH *path = gdgpath_new ( partall );
    path->dst_net = net;
    path->create_path_str = create_path_str;
    path->verbose = verbose;

    if ( !gdgpath_add_net ( path, net ) ) {
        gdgpath_destroy ( path );
        return NULL;
    };

    return path;
}


char* gdgpath_get_elements_txt ( st_GDGPATH *path ) {
    GString *str = g_string_new ( "NENI NIC!" );

    return g_string_free ( str, FALSE );
}


char* gdgpath_get_vhdl ( st_GDGPATH *path, gboolean optimalized_output ) {

    st_GDGELEMENTSALL *elmall = gdgelements_create_slave_from_elements ( path->element, path->element_count );

    st_GDGPARTSALL *partall = gdgparts_create_slave_from_parts ( path->part, path->part_count, elmall );
    if ( !partall ) return NULL;

    int i;
    for ( i = 0; i < partall->count_part; i++ ) {
        gdgelement_register_part_instance ( partall->elmall, partall->part[i]->element, partall->part[i] );
    };

    st_GDGNETSALL *netall = gdgnets_create_slave_from_nets ( path->net, path->net_count );

    st_GDGIOPORTSALL *ioportall = gdgnets_create_ioports ( netall, partall );

    GString *str = g_string_new ( "" );
    g_string_append_printf ( str, "GDG_module_%s", path->dst_net->name );
    char *vhdl_txt = vhdl_create ( str->str, partall, netall, ioportall, optimalized_output );
    g_string_free ( str, TRUE );

    gdgnets_ioports_destroy ( ioportall );
    gdgparts_destroy ( partall );
    gdgnets_destroy ( netall );
    gdgelements_destroy ( elmall );

    return vhdl_txt;
}
